#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>


/*
  mkspifs

  Usage: mkspifs -o image.img file1 [file2 [file3]*]

  Creates an image file with the custom file system structure
  that is used by the spiplayer.

  Adds the file listed on the command like in that order.

  The display name of each file is added into the reserved 12
  bytes in the directory entry.

 */

typedef unsigned char u_char;
typedef unsigned long u_long;

#include "spifs.h"

u_long GetULong(u_char *sector, int offset) {
    return sector[offset] | ((u_long)sector[offset+1]<<8) |
	((u_long)sector[offset+2]<<16) |
	((u_long)sector[offset+3]<<24);
}
u_long GetUShort(u_char *sector, int offset) {
    return sector[offset] | ((u_long)sector[offset+1]<<8);
}
void PutULong(u_char *sector, int offset, u_long value) {
    sector[offset+0] = value;
    sector[offset+1] = value>>8;
    sector[offset+2] = value>>16;
    sector[offset+3] = value>>24;
}
void PutUShort(u_char *sector, int offset, u_long value) {
    sector[offset+0] = value;
    sector[offset+1] = value>>8;
}


#define F_ERROR 1
int flags = 0;

u_char *headerData = NULL;
u_long headerSize = 0;
u_long dirEntriesAllocated = 0, dirEntriesUsed = 0, numberOfFiles = 0;
u_char *fileData = NULL;
u_long fileDataSize = 0;


void AddFile(const char *name, const char *displayName) {
    int index, i;
    FILE *fp;
    const char *nm = name;

    /* strip away path, if any (mainly the list file can have a path) */
    if ((nm = strrchr(name, '\\'))) {
	nm++;
    } else if ((nm = strrchr(name, '/'))) {
	nm++;
    } else {
	nm = name;
    }
    /* If no directory entries left, make space */
    if (dirEntriesUsed >= dirEntriesAllocated) {
	u_char *new = realloc(headerData, headerSize+512);
	if (new) {
	    headerData = new;
	    memset(headerData+headerSize, 0, 512); /* clear the new sector */
	    headerSize += 512;
	    dirEntriesAllocated += SPIFILEENTRIESPERBLOCK;
	} else {
	    fprintf(stderr, "Allocation failed for %ld bytes\n",
		    headerSize+512);
	    exit(EXIT_FAILURE); /*performs free for headerData and fileData*/
	}
    }
    /* We now have space for this directory entry */
    if ((fp = fopen(name, "rb"))) {
	u_long fileSize;

	/*determine file size*/
	fseek(fp, 0, SEEK_END);
	fileSize = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	printf("Adding File '%s' '%s' %ld bytes\n", name, displayName, fileSize);

	/* location of the new entry */
	index = 12 + (dirEntriesUsed/SPIFILEENTRIESPERBLOCK) * 512 +
	    (dirEntriesUsed % SPIFILEENTRIESPERBLOCK)
	    * sizeof(struct SPIFILEENTRY);

	/*preset name with spaces*/
	memset(headerData+index+0, ' ', 12);
	/* put filename, convert to upper case */
	for (i=0;i<8;i++) {
	    if (nm[i] == '.') /*premature separator*/
		break;
	    headerData[index+0+i] = toupper(nm[i]);
	}
	if (nm[i] == '.') {
	    i++;
	}
	/* put suffix */
	headerData[index+8] = toupper(nm[i+0]);
	headerData[index+9] = toupper(nm[i+1]);
	headerData[index+10] = toupper(nm[i+2]);
	/* put display name into the dmy[12] field */
	//memcpy(&headerData[index+20], displayName, 12);

	/*start sector, relative to data start now (until we know the
	  start of the data blocks).*/
	PutULong(headerData, index+12, fileDataSize/512);
	PutULong(headerData, index+16, fileSize); /*used by ogg player*/
	/* Allocate space for the file */
	{
	    u_char *new = realloc(fileData,
				  fileDataSize + ((fileSize+511)&~511L));
	    if (!new) {
		fprintf(stderr, "Allocation2 failed for %ld bytes\n",
			fileDataSize + ((fileSize+511)&~511L));
		exit(EXIT_FAILURE);
	    }
	    fileData = new;
	}
	/* Then copy the data, one sector at a time */
	while (fileSize) {
	    u_char sector[512];
	    memset(sector, 0, sizeof(sector)); /*clear previous data*/
	    fread(sector, 512, 1, fp);
	    /* Add the sector to the data blocks */
	    memcpy(fileData+fileDataSize, sector, 512);
	    fileDataSize += 512;
	    if (fileSize < 512) {
		fileSize = 0;
	    } else {
		fileSize -= 512;
	    }
	}
	fclose(fp);

	dirEntriesUsed++;
	numberOfFiles++;
    } else {
	fprintf(stderr, "Could not open '%s' for reading, skipping\n", name);
    }
}

int main(int argc, char *argv[]) {
    int i, n;
    //int numFiles = 0;
    const char *listFile = NULL;
    const char *imgFile = NULL;
    FILE *fp, *fpo;

    /* Start with header for 50 files */
    dirEntriesUsed = 0;
    dirEntriesAllocated = 2*SPIFILEENTRIESPERBLOCK;
    headerSize = 2*512;
    headerData = calloc(headerSize, 1);

    for (n=1; n<argc; n++) {
	if (argv[n][0] == '-') {
	    /*options*/
	    if (argv[n][1] == 'o') { /* set output file (image) */
		if (!argv[n][2] && n+1 < argc) {
		    imgFile = argv[++n];
		} else {
		    imgFile = &argv[n][2];
		}
	    } else {
		/* other options */
		flags |= F_ERROR;
	    }
	} else {
	    /* a file name, add that to image */
	    if (listFile)
		flags |= F_ERROR; /* only allow one file */
	    listFile = argv[n];
	}
    }
    if (!listFile || (flags & F_ERROR)) {
	fprintf(stderr,
		"Usage: %s [-<options>] [-o <imagefile>] <listfile>\n",
		argv[0]);
	return 10;
    }
    if ((fp = fopen(listFile, "rb"))) {
	char line[200], lnum = 1;
	while (fgets(line, 200-1, fp)) {
	    u_char *p;
//	    printf("%s", line);
	    p = line;
	    while (*p && isspace(*p))
		p++;
	    if (*p == '#')
		continue;
	    p = strchr(line, ','); /* locate the comma */
	    if (p) {
		u_char *n;
		*p++ = '\0';
		n = p;
		/* all characters upto newline or carriage return */
		while (*n) {
		    if (*n == '\n' || *n == '\r') {
			*n = '\0';
			break;
		    }
		    n++;
		}
		/* add file with the specified display-name */
		AddFile(line, p);
	    } else {
		p = line;
		while (*p) {
		    if (*p == '\n' || *p == '\r') {
			*p = '\0';
			break;
		    }
		    p++;
		}
		AddFile(line, "");
	    }
	    lnum++;
	}
	fclose(fp);
    } else {
	fprintf(stderr, "Could not open list file '%s' for reading\n", listFile);
	return 20;
    }
    printf("%ld files (%ld)\n", numberOfFiles, dirEntriesUsed);
    printf("Header %ld bytes (%ld sectors)\n", headerSize, headerSize/512);
    printf("Data   %ld bytes (%ld sectors)\n", fileDataSize, fileDataSize/512);
    /* Can now fill the rest of the header fields */
    PutULong(headerData, 0x004, headerSize/512); /*number of directory sectors*/
    /*
      The "sec" field is relative to data start.
    */

    /* Write the collected image. */
    if (imgFile) {
	if ((fpo = fopen(imgFile, "wb"))) {
	    fwrite(headerData, headerSize, 1, fpo);
	    fwrite(fileData, fileDataSize, 1, fpo);
	    fclose(fpo);
	} else {
	    fprintf(stderr, "Could not open '%s' for writing\n", imgFile);
	}
    } else {
	fprintf(stderr, "No image file specified with -o option\n");
	return 5;
    }
    return 0;
}
