#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "spifs.h"
#include "modprogimg.h" /* VS1000 programmer */

int reservedSectors = SPIFSSTARTBLOCK;


/***********************************************************************/
unsigned long GetULong(unsigned char *sector, int offset) {
    return sector[offset] | ((unsigned long)sector[offset+1]<<8) |
	((unsigned long)sector[offset+2]<<16) |
	((unsigned long)sector[offset+3]<<24);
}
unsigned long GetUShort(unsigned char *sector, int offset) {
    return sector[offset] | ((unsigned long)sector[offset+1]<<8);
}
void PutULong(unsigned char *sector, int offset, unsigned long value) {
    sector[offset+0] = value;
    sector[offset+1] = value>>8;
    sector[offset+2] = value>>16;
    sector[offset+3] = value>>24;
}
void PutUShort(unsigned char *sector, int offset, unsigned long value) {
    sector[offset+0] = value;
    sector[offset+1] = value>>8;
}
unsigned char *headerData = NULL;
unsigned long headerSize = 0;
unsigned long dirEntriesAllocated = 0, dirEntriesUsed = 0, numberOfFiles = 0;
unsigned char *fileData = NULL;
unsigned long fileDataSize = 0;

void AddFile(const char *name, const char *displayName) {
    int index, i;
    FILE *fp;
    const char *nm = name;

    /* strip away path, if any (mainly the list file can have a path) */
    if ((nm = strrchr(name, '\\'))) {
	nm++;
    } else if ((nm = strrchr(name, '/'))) {
	nm++;
    } else {
	nm = name;
    }
    /* If no directory entries left, make space */
    if (dirEntriesUsed >= dirEntriesAllocated) {
	unsigned char *new = realloc(headerData, headerSize+512);
	if (new) {
	    headerData = new;
	    memset(headerData+headerSize, -1, 512); /* 'clear' the new sector */
	    headerSize += 512;
	    dirEntriesAllocated += SPIFILEENTRIESPERBLOCK;
	} else {
	    fprintf(stderr, "Allocation failed for %ld bytes\n",
		    headerSize+512);
	    exit(EXIT_FAILURE); /*performs free for headerData and fileData*/
	}
    }
    /* We now have space for this directory entry */
    if ((fp = fopen(name, "rb"))) {
	unsigned long fileSize;

	/*determine file size*/
	fseek(fp, 0, SEEK_END);
	fileSize = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	printf("Adding File '%s' '%s' %ld bytes\n", name, displayName, fileSize);

	/* location of the new entry */
	index = 12 + (dirEntriesUsed/SPIFILEENTRIESPERBLOCK) * 512 +
	    (dirEntriesUsed % SPIFILEENTRIESPERBLOCK)
	    * sizeof(struct SPIFILEENTRY);

	/*preset name with spaces*/
	memset(headerData+index+0, ' ', 12);
	/* put filename, convert to upper case */
	for (i=0;i<8;i++) {
	    if (nm[i] == '.') /*premature separator*/
		break;
	    headerData[index+0+i] = toupper(nm[i]);
	}
	if (nm[i] == '.') {
	    i++;
	}
	/* put suffix */
	headerData[index+8] = toupper(nm[i+0]);
	headerData[index+9] = toupper(nm[i+1]);
	headerData[index+10] = toupper(nm[i+2]);

	/*start sector, relative to data start now (until we know the
	  start of the data blocks).*/
	PutULong(headerData, index+12, fileDataSize/512);
	PutULong(headerData, index+16, fileSize); /*used by ogg player*/
	/* Allocate space for the file */
	{
	    unsigned char *new = realloc(fileData,
				  fileDataSize + ((fileSize+511)&~511L));
	    if (!new) {
		fprintf(stderr, "Allocation2 failed for %ld bytes\n",
			fileDataSize + ((fileSize+511)&~511L));
		exit(EXIT_FAILURE);
	    }
	    fileData = new;
	}
	/* Then copy the data, one sector at a time */
	while (fileSize) {
	    unsigned char sector[512];
	    memset(sector, 0, sizeof(sector)); /*clear previous data*/
	    fread(sector, 512, 1, fp);
	    /* Add the sector to the data blocks */
	    memcpy(fileData+fileDataSize, sector, 512);
	    fileDataSize += 512;
	    if (fileSize < 512) {
		fileSize = 0;
	    } else {
		fileSize -= 512;
	    }
	}
	fclose(fp);

	dirEntriesUsed++;
	numberOfFiles++;
    } else {
	fprintf(stderr, "Could not open '%s' for reading, skipping\n", name);
    }
}

void InitSpiFs(void) {
    /* Start with header for 50 files */
    dirEntriesUsed = 0;
    dirEntriesAllocated = 2*SPIFILEENTRIESPERBLOCK;
    headerSize = 2*512;
    headerData = calloc(headerSize, 1);
    memset(headerData, -1, headerSize);
}
/***************************************************************************/




/***************************************************************************/
int RS_InitComm(int comPort);
int RS_CloseComm(void);
int RS_SetCommSpeed(long speed);
int RS_SendData(char *data, int len);
int RS_ReceiveData(char *data, int len);
void RS_Wait(void);
void RS_Drain(void);

/*

FILE: rs232.c - serial connection functions
DESCRIPTION: Serial communication for unix / DOS

*/
#define BAUDRATE B9600
#define DEF_BAUDRATE 9600

unsigned int baudRate=DEF_BAUDRATE;
volatile int * volatile breakPtr = NULL;

void RS_SetBreakIntVar(volatile int *breakVar) {
    breakPtr = breakVar;
}

#ifndef WIN32

#include <sys/types.h>
#include <termios.h>
#include <fcntl.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* FreeBSD termios-based serial communication */

int newsfd;
struct termios oldtio;

void RS_Wait(void) {
    usleep(100000);
}

void RS_Drain(void) {
    tcdrain(newsfd);
}

int RS_InitComm(int comPort) {
    static char commDevice[80];
#if !defined(__linux__) && !defined(linux)
    sprintf(commDevice, "/dev/cuaa%d", comPort-1);
#else
    sprintf(commDevice, "/dev/ttyS%d", comPort-1);
#endif
    /* support ttyUSBxx */
    if (comPort >= 1000) {
	sprintf(commDevice, "/dev/ttyUSB%d", comPort - 1000);
    }
    newsfd = open(commDevice, O_RDWR | O_NOCTTY);
    if (newsfd <0) {
	perror(commDevice);
	return -1;
    }
    tcgetattr(newsfd,&oldtio); /* save current port settings */
    return RS_SetCommSpeed(baudRate);
}

void RS_SendBreak() {
    tcsendbreak(newsfd, 0); /* 0.25 .. 0.5 sec break */
}

int RS_SendData(char *data, int len) {
    if (write(newsfd, data, len) < 0)
	perror("Write error");
    return 0;
}

int RS_ReceiveData(char *data, int len) {
  int i = 0;

  while (len - i > 0) {
      int k = read(newsfd, data+i, len - i);

//fprintf(stderr, "Read %d\n", k);
      if (k < 0) {
	  /* read error, restart read */
//fprintf(stderr, "Read err, restart\n");
      } else {
	  i += k;
	  /* timeout, check breakPtr */
	  if (k == 0 && breakPtr) {
	      if (breakPtr && *breakPtr) {
//fprintf(stderr, "timeout + break\n");
		  return -(i+1);
	      }
	  }
      }
  }
  return 0;
}

int RS_CloseComm(void) {
    /* restore old port settings */
    tcsetattr(newsfd,TCSANOW,&oldtio);
    /* close serial port */
    close(newsfd);
    return 0;
}

#if defined(__linux__) || defined(linux)
#include <signal.h>
#include <sys/resource.h>
#include <linux/serial.h>
#include <sys/ioctl.h>
#endif



int Set_DTR(void) {
    int nRet;
    int status;
    status = 0;
    nRet = ioctl(newsfd, TIOCMGET,&status);
    if (nRet < 0)
	return 0;
    status |= TIOCM_DTR;
    nRet = ioctl(newsfd, TIOCMSET, &status);
    if (nRet < 0)
	return 0;
    return 1;
}


int Clr_DTR(void) {
    int nRet;
    int status;
    status = 0;
    nRet = ioctl(newsfd, TIOCMGET,&status);
    if (nRet < 0)
	return 0;
    status &= ~TIOCM_DTR;
    nRet = ioctl(newsfd, TIOCMSET, &status);
    if (nRet < 0)
	return 0;
    return 1;
}


struct Speed {
  unsigned long symbol;
  int bps;
} speeds[] = {
  {B0      , 	  0},	/* hang up */
  {B50     ,      50},
  {B75     ,      75},
  {B110    ,     110},
  {B134    ,     134},
  {B150    ,     150},
  {B200    ,     200},
  {B300    ,     300},
  {B600    ,     600},
  {B1200   ,    1200},
  {B1800   ,    1800},
  {B2400   ,    2400},
  {B4800   ,    4800},
  {B9600   ,    9600},
  {B19200  ,   19200},
  {B38400  ,   38400},
  {B57600  ,   57600},
  {B115200 ,  115200},
  {B230400 ,  230400},
#ifdef B460800
  {B460800 ,  460800},
#endif
#ifdef B500000
  {B500000 ,  500000},
#endif
#ifdef B576000
  {B576000 ,  576000},
#endif
#ifdef B921600
  {B921600 ,  921600},
#endif
#ifdef B1000000
  {B1000000, 1000000},
#endif
#ifdef B1152000
  {B1152000, 1152000},
#endif
#ifdef B1500000
  {B1500000, 1500000},
#endif
#ifdef B2000000
  {B2000000, 2000000},
#endif
#ifdef B2500000
  {B2500000, 2500000},
#endif
#ifdef B3000000
  {B3000000, 3000000},
#endif
#ifdef B3500000
  {B3500000, 3500000},
#endif
#ifdef B4000000
  {B4000000, 4000000},
#endif
};

/*

FUNCTION: SetCommSpeed()
DESCRIPTION: Sets communications speed
INPUTS: speed - desired speed in bps
RESULT: 0 if there were no errors, -1 otherwise
SEE ALSO:

*/
int RS_SetCommSpeed(long speed) {
    struct termios speedtio;
    unsigned long speedSymbol;
    int i = 0;
#if defined(__linux__) || defined(linux)
    int customSpeed = 0;
#endif

    while (i < sizeof(speeds)/sizeof(speeds[0]) && speed != speeds[i].bps)
	i++;

    if (i >= sizeof(speeds)/sizeof(speeds[0])) {
#if defined(__linux__) || defined(linux)
	customSpeed = 1;
	speedSymbol = B38400;
#else
	speedSymbol = speed; /*FreeBSD takes a value instead of bit mask!*/
#endif
    } else {
	speedSymbol = speeds[i].symbol;
    }

    /* set new port settings for canonical input processing */
    memset(&speedtio, 0, sizeof(speedtio));
    tcgetattr(newsfd,&speedtio); /* save current port settings */
    speedtio.c_cflag = CS8 | CLOCAL | CREAD ;
    speedtio.c_iflag = IGNPAR;
    speedtio.c_oflag = 0;
    speedtio.c_lflag = 0;

    cfmakeraw(&speedtio);

    speedtio.c_cc[VMIN] = 0;
    speedtio.c_cc[VTIME] = 30;//15; /* in 0.1sec granularity, VMIN=0 */

#if defined(__linux__) || defined(linux)
    if (customSpeed) {
	struct serial_struct ss;
	int closestSpeed;

	ioctl(newsfd, TIOCGSERIAL, &ss);
	ss.flags = (ss.flags & ~ASYNC_SPD_MASK) | ASYNC_SPD_CUST;
	ss.custom_divisor = (ss.baud_base + (speed / 2)) / speed;
	if (ss.custom_divisor == 0)
	    ss.custom_divisor = 1;
	closestSpeed = ss.baud_base / ss.custom_divisor;

	if (closestSpeed < speed * 98 / 100 ||
	    closestSpeed > speed * 102 / 100) {
	    fprintf(stderr, "Cannot set serial port speed to %ld. "
		    "Closest possible is %d\n",
		    speed, closestSpeed);
	}
	ioctl(newsfd, TIOCSSERIAL, &ss);
    }
#endif /*defined(__linux__) || defined(linux)*/

    cfsetispeed(&speedtio, speedSymbol);
    cfsetospeed(&speedtio, speedSymbol);
    if (tcsetattr(newsfd, TCSADRAIN, &speedtio)==-1) {
	printf("Serial port input speed (%ld) setting failed, errno: %d\n",
	       speed, errno);
	return -1;
    }
//   Set_DTR();
    Clr_DTR();
    return 0;
}

#else

/* See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/devio/base/about_communications_resources.asp */
#include <windows.h>

#ifdef WIN32
#define ErrorPrint(x) fprintf(stderr,x)
#endif 
/* module variables*/
HANDLE       SerialHandle;
DCB          Control;	      /* device control block */
COMMTIMEOUTS TimeOuts;

int RS_InitComm(int comPort)  {
    static char portName[80];
    if (comPort > 9) {
	sprintf(portName, "\\\\.\\COM%d", comPort);
    } else {
	sprintf(portName, "COM%d", comPort);
    }
    SerialHandle = CreateFile(portName,
			      GENERIC_READ|GENERIC_WRITE,
			      0,                  /* not shared */
			      NULL,		 /* no security attributes*/
			      OPEN_EXISTING,
			      FILE_ATTRIBUTE_NORMAL|FILE_FLAG_NO_BUFFERING|
			      FILE_FLAG_WRITE_THROUGH,
			      NULL);
  
    if (SerialHandle == INVALID_HANDLE_VALUE) {
	ErrorPrint("Couldn't open serial port!\n");
	return -1;
    }
    if (!BuildCommDCB("baud=1200 parity=N data=8 stop=1", &Control)) {
	ErrorPrint("Couldn't build control block!\n");
	return -1;
    }
    Control.DCBlength = sizeof(DCB);
    Control.BaudRate = baudRate;
    Control.fBinary = TRUE;
    Control.fParity = FALSE;
    Control.fOutxCtsFlow = FALSE;
    Control.fOutxDsrFlow = FALSE;
    Control.fDtrControl = DTR_CONTROL_DISABLE;
    Control.fDsrSensitivity = FALSE;
    Control.fTXContinueOnXoff = FALSE;
    Control.fOutX = FALSE;
    Control.fInX = FALSE;
    Control.fErrorChar = FALSE;
    Control.fNull = FALSE;
    Control.fRtsControl = RTS_CONTROL_DISABLE;
    Control.fAbortOnError = TRUE;
    /* set user defined values */
    if (!SetCommState(SerialHandle, &Control)) {
	ErrorPrint("Couldn't set control state!\n");
	return -1;
    }

    TimeOuts.ReadIntervalTimeout    = 0; /* max. time between two bytes in ms*/
    TimeOuts.ReadTotalTimeoutMultiplier = 0; /*1second/byte allowed*/
    TimeOuts.ReadTotalTimeoutConstant   = 4000; /*ms overhead in the start of tx*/
    TimeOuts.WriteTotalTimeoutMultiplier= 0; /*ms/write allowed */
    TimeOuts.WriteTotalTimeoutConstant  = 0; /*ms overhead in the start of tx*/

    if (!SetCommTimeouts(SerialHandle, &TimeOuts)) {
	ErrorPrint("Couldn't set timeout values!\n");
	return -1;
    }
    /* clear input buffer */
    PurgeComm(SerialHandle, PURGE_RXABORT | PURGE_RXCLEAR);
    return 0;
}

int RS_ReceiveData(char *data, int len) {
    int i = 0;

    while (len - i > 0) {
	DWORD read = 0;

	ReadFile(SerialHandle, (BYTE *)data+i, len-i, &read, NULL);
	if (read < 0) {
	    /* read error */
	} else {
	    i += read;
	    /* timeout, check breakPtr */
	    if (read == 0 && breakPtr) {
		if (breakPtr && *breakPtr)
		    return -(i+1);
	    }
	}
    }
    return 0;
}

void RS_SendBreak() {
    /*tcsendbreak(newsfd, 0);*/ /* 0.25 .. 0.5 sec break */
    RS_Drain();
    SetCommBreak(SerialHandle); /*EscapeCommFunction(SerialHandle, SETBREAK);*/
    Sleep(500);
    ClearCommBreak(SerialHandle); /*EscapeCommFunction(SerialHandle, CLRBREAK);*/
}

int RS_SendData(char *data, int len) {
    DWORD written;
  
    WriteFile(SerialHandle, (BYTE *)data, len, &written, NULL);
    if (written != len) {
	ErrorPrint("Write byte failed!\n");
	return -1;
    }
	return 0;
}

int RS_CloseComm(void) {
    PurgeComm(SerialHandle, PURGE_RXABORT | PURGE_RXCLEAR);
    CloseHandle(SerialHandle);
    return 0; 
} 
 
int RS_SetCommSpeed(long speed) {
    if (!GetCommState(SerialHandle, &Control)) {
	ErrorPrint("Couldn't get control state!\n");
	return -1;
    }
    Control.BaudRate = speed;
    if (!SetCommState(SerialHandle, &Control)) {
	ErrorPrint("Couldn't set control state!\n");
	return -1;
    }
    return 0;
}

void RS_Drain(void) {
    FlushFileBuffers(SerialHandle);
}

void RS_Wait(void) {
    Sleep(300);
}
#endif
/***************************************************************************/


#define RS232_SPEED 9600


/***************************************************************************/
#define BLSIZE (512) //512 bytes
int fileSize = 0;
int blSize = -1;
char block[BLSIZE+10];

void GetBlock(FILE *fp) {
    memset(block+6, 0, BLSIZE);
    blSize = fread(block+6, 1, BLSIZE, fp);
    fileSize += blSize;
    if (blSize & 1)
	blSize++;
}
void SendBlock(int addr, int size /*1=ok,0=eof,255=eop*/) {
    int i, ch = 0xffff;
//printf("Sending %d, size %d\n", addr, size);
    block[0] = 0xbb; /*Start marker*/
    block[1] = size; /*Size in blocks, 0=end of file, 255=end of programming */
    block[2] = addr>>8;
    block[3] = addr;

    for (i=0;i<512/2;i++) {
	ch += (block[6+2*i]<<8) + (block[7+2*i]&0xff);
    }
    ch = ~ch;
//printf("%04x\n", ch &0xffff);
    block[4] = ch>>8;
    block[5] = ch;
    RS_SendData(block, 512+6);
}


void putch(int c) {
    unsigned char ch = c;
    RS_SendData(&ch, 1);
}
void putword(unsigned short w) {
    unsigned char ch[2];
    ch[0] = w>>8;
    ch[1] = w;
    RS_SendData(ch, 2);
}


long sentblocks = 0;
int sizeBits = 0;
int SendFile(const char *name, int startBlock, int cont, int last) {
    FILE *fp;
    int addr = startBlock-1;
    unsigned char o = 0;
//    volatile int breakVar = 1;

    fp = name ? fopen(name, "rb") : NULL;
    if (!fp) {
	printf("Could not open '%s' for reading\n", name ? name : "");
	exit(EXIT_FAILURE);
    }
//    RS_SetBreakIntVar(&breakVar); //Allow timeouts
    while (1) {
	unsigned char d;
	int c;

	if (cont) {
//	    fprintf(stderr, "Continuing with another file..\n");
	    o = ('' & 255);
	    d = 0x42;
	    blSize = -1;
	    goto doit;
	}
	if (!RS_ReceiveData((char *)&d, 1)) {
	doit:
	    cont = 0;
	    c = d;
//fprintf(stderr, "%c %02x\n", isprint(c) ? c : ' ', c);
//fprintf(stderr, "%02x ", c);
	    if (o != ('' & 255)) {
		//printf(" %02x<-%02x\n", o, d);
		o = d;
		continue;
	    }
	    o = d;
	    //printf(" old <- %d, c = %d\n", o, c);
	    if (c == 0x42) {
		if (!blSize) {
		    fclose(fp);
		    fp = NULL;
		    break;
		}
		fprintf(stderr, "\r%4d bytes block %d ", fileSize, addr);
		addr++;
		GetBlock(fp);

		SendBlock(addr, blSize ? 1 : (last ? 255 : 0));
		sentblocks++;
	    } else if (c == 0x43) {
		fprintf(stderr, "\nUART Transfer Failed, Retrying..\n");
		SendBlock(addr, blSize ? 1 : (last ? 255 : 0));
	    } else if (c == 0x44) {
		fprintf(stderr, "\nSPI Programming Failed at %d!\n", addr-1);
		return 1;
	    } else if (c == 0x45) {
		fprintf(stderr, "\nSPI Programming Successful!\n");
		return 0;
	    } else if (c == 0x40) {
		if (sizeBits >= 20) {
		    fprintf(stderr, "%dMB\n", (1<<(sizeBits-20)));
		}
		fprintf(stderr, "VS1103 Ready, performing chip erase.\n");
	    } else if (c == 0x00 || c == 0xff) {
	      fprintf(stderr, "Unknown size, REMS not supported, or solder issues (0x%02x).\n", c);
	    } else if (c >= 0x15 && c <= 0x18) {
		if (sizeBits) {
		    if (sizeBits != c) {
			fprintf(stderr, "Reported size 0x%02x does not match with specified size 0x%02x!\n", c, sizeBits);
			return 1;
		    }
		} else {
		    sizeBits = c;
		}
	    } else if (c == 0x41) {
		fprintf(stderr, "Chip Erase complete.\n");
	    } else if (c == 0x46) {
		fprintf(stderr, "Chip Erase skipped.\n");
	    } else {
		fprintf(stderr, "Unknown response %c (0x%02x)\n",
			isprint(c) ? c : ' ', c);
	    }
	} else {
	    /* Timeout -- should wait until ~30 sec. */
	    //fprintf(stderr, "\nTimeout before starting!\n");
	    //return 1;
	}
    }
    fprintf(stderr, "\n");
//    RS_SetBreakIntVar(NULL); //no timeouts
    return 0;
}

#include <time.h>
int main(int argc, char *argv[]) {
    FILE *fp = NULL;
    const char *inName = NULL;
    const char *imgName = NULL;
    int port = 1, n, serMul = 1;
    time_t stime = time(NULL), etime;
    const char *imgFile = "tmp.img";
    FILE *fpo = NULL;

    baudRate = RS232_SPEED;
    InitSpiFs();
    for (n=1; n<argc; n++) {
	if (!strncmp(argv[n], "-r", 2) && argv[n][2]) {
	    reservedSectors = strtol(argv[n]+2, NULL, 0);
	} else if (!strcmp(argv[n], "-f") && n+1<argc) {
	    const char *name = argv[++n];
	    AddFile(name, name);
	} else if (!strcmp(argv[n], "-p") && n+1<argc) {
	    port = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-s") && n+1<argc) {
	    baudRate = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-sm") && n+1<argc) {
	    serMul = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-b") && n+1<argc) {
	    sizeBits = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-l") && n+1<argc) {
	    if (inName) {
		fprintf(stderr, "Warning: only one input file allowed, "
			"%s discarded\n", inName);
	    }
	    inName = argv[++n];
	} else if (!strcmp(argv[n], "-i") && n+1<argc) {
	    if (imgName) {
		fprintf(stderr, "Warning: only one content image file allowed, "
			"%s discarded\n", imgName);
	    }
	    imgName = argv[++n];
	} else {
	  fprintf(stderr,
		  "Usage: %s [-p <port>] [-s <speed>] [-l <firmwarefile>]\n"
		  "      [-i <fatimage>] [-r <reservedsectors>]  Default: 64\n"
		  "      [-sm <sermul>] [-b <sizebits>]\n"
		  "Example: pcmodflash -p 1 -s 9600 -r 64 -l firmware.spi -i content.img\n"
		  "       -p 1000 for ttyUSB0 etc.\n"
		  "       -f <file> to add a file directly\n"
		  "Example: pcmodflash -p 1 -s 9600 -r 64 -l firmware.spi -f 1khz.wav -f bell.wav\n"
		  ,
		  argv[0]);
	  exit(EXIT_FAILURE);
	}
    }

    if (numberOfFiles) {
	if (imgName) {
	    fprintf(stderr, "Error: "
		    "can not specify both image file and separate files\n");
	    return 5;
	}
	printf("%ld files (%ld)\n", numberOfFiles, dirEntriesUsed);
	printf("Header %ld bytes (%ld sectors)\n", headerSize, headerSize/512);
	printf("Data   %ld bytes (%ld sectors)\n", fileDataSize, fileDataSize/512);
	/* Can now fill the rest of the header fields */
	/*number of directory sectors*/
	PutULong(headerData, 0x004, headerSize/512);
	if (imgFile) {
	    if ((fpo = fopen(imgFile, "wb"))) {
		fwrite(headerData, headerSize, 1, fpo);
		fwrite(fileData, fileDataSize, 1, fpo);
		fclose(fpo);
	    } else {
		fprintf(stderr, "Could not open '%s' for writing\n", imgFile);
	    }
	} else {
	    fprintf(stderr, "No image file specified with -o option\n");
	    return 5;
	}
	imgName = imgFile;
    }

    if (RS_InitComm(port)) {
	fprintf(stderr, "Could not open serial!\n");
	exit(-1);
    }
    if (RS_SetCommSpeed(baudRate)) {
	RS_CloseComm();
	return 10;
    }

    fprintf(stderr, "Trying to connect..\n");
    /*Wake up monitor*/
    RS_SendData("\xef", 1);
    {
	volatile int breakVar = 1;
	int tries = 8;
	RS_SetBreakIntVar(&breakVar);
	while (1) {
	    unsigned char d;
	    int c;
	    if (!RS_ReceiveData((char *)&d, 1)) {
		c = d;
		fprintf(stderr, "%02x ", c);
		if (c == 0xa5) {
		    fprintf(stderr, "VS10XX\n");
		    break; /*VS1053*/
		}
		if (c == 0xfe) {
		    /* VS1000 */
		    fprintf(stderr, "VS1063 (VS1000 not supported by this)\n");
		    break;/*Accept also 0xfe as a response to "jump to monitor" command.*/
		}
	    } else if (--tries > 0) {
		fprintf(stderr, "Trying again.\n");
		RS_SendData("\xef", 1);
	    } else {
		fprintf(stderr, "Timeout, giving up. Check power and cables.\n");
		RS_CloseComm();
		return 10;
	    }
	}
	RS_SetBreakIntVar(NULL);
    }


    /* Save serMul into vs1103 register 0xc00c */
    putch(0x72);
    putword(1-1); //1 word
    putword(0);   //addr hi
    putword(0xc00c); //addr lo
    putch(0<<2); //0=write, 0<<2 = XMEM, 1<<2 = YMEM
    putword(serMul);


    printf("Loading code...\n");
    /* load+start software */
    {
	unsigned short *p = &autoprog[0];
	unsigned short head[3];
	int j;
	while (1) {
	    memcpy(head, p, sizeof(head));
	    p += 3;

	    if (head[0] == 0x8000U) {
		//printf("I\n");

		putch(0x72); /*'r'*/
		putword(head[1]>>1); //len-1
		putword(0);          //addr hi
		putword(head[2]);   //addr lo
		putch(2<<2); //0=write, 2<<2 = IMEM

		head[1] += 1;
		while (head[1] > 0) {
		    register int words = head[1];
		    if (words > 256)
			words = 256;
		    head[1] -= words;

		    for (j=0;j<words;j++) {
			putword(*p++);
		    }
		}
	    } else if (head[0] == 0x8001U || head[0] == 0x8002U) {
		//printf(head[0] == 0x8001U ? "X\n" : "Y\n");

		putch(0x72);
		putword(head[1]);
		putword(0);
		putword(head[2]);
		putch((head[0]-1)<<2); //0=write, 0<<2 = XMEM, 1<<2 = YMEM

		head[1] += 1;
		while (head[1] > 0) {
		    register int words = head[1];
		    if (words > 256)
			words = 256;
		    head[1] -= words;
		    for (j=0;j<words;j++) {
			putword(*p++);
		    }
		}
		if (head[0] == 0x8002U) {
		    /*Because of the bug..skip a word*/
		    p+=1;
		}

	    } else if (head[0] == 0x8003U) {
		printf("Executing VS1103/VS10xx code (0x%04x)\n", head[2]);

		putch('e'); /*exec*/
		putword(0);
		putword(head[2]);
		putword(0x0010); /*emuarea*/
		putword(0x0000); /*nop*/
		putword(0x0024);
		break;
	    } else {
		break;
	    }
	}
    }
	RS_Drain();
	RS_Wait();
//	RS_Wait();
#if 1
    if (RS_SetCommSpeed(115200*serMul)) {
	RS_CloseComm();
	return 10;
    }
    fprintf(stderr, "Changed UART speed to %ldbps.\n", 115200L * serMul);
#endif
	RS_Wait();
//	RS_Wait();
	RS_Drain();
//	RS_Wait();
	RS_Wait();

    /* Program content image first, then the firmware */
    if (imgName) {
	fprintf(stderr, "Programming %s at sector %d\n",
		imgName, reservedSectors);
	if (SendFile(imgName, reservedSectors, 0, inName?0:1/*last*/)) {
	    fprintf(stderr, "Programming failed\n");
	    RS_CloseComm();
	    return 10;
	}
    }
    if (inName) {
	fprintf(stderr, "Programming %s\n", inName);
	if (SendFile(inName, 0, imgName ? 1 : 0, 1/*last*/)) {
	    fprintf(stderr, "Programming failed\n");
	    RS_CloseComm();
	    return 10;
	}
    }
    fprintf(stderr, "Programming Finished Successfully\n");
    RS_CloseComm();
    etime = time(NULL);
    fprintf(stderr, "%ld blocks programmed in %d seconds, %5.2fkB/sec\n",
	    sentblocks, (int)(etime-stime), sentblocks*0.5/(etime-stime));
    return 0;
}
